import { Promise } from "./promise";
import { resolve } from "./resolve";

// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Error
function ResolvedQueueError(message) {
  this.name = "resolved";
  if ((message !== undefined) && (typeof message !== "string")) {
    throw new TypeError('You must pass a string.');
  }
  this.message = message || "Default Message";
}
ResolvedQueueError.prototype = new Error();
ResolvedQueueError.prototype.constructor = ResolvedQueueError;

var Queue = function() {
  var queue = this,
    promise_list = [],
    promise,
    fulfill,
    reject,
    resolved;

  if (!(this instanceof Queue)) {
    return new Queue();
  }

  function canceller() {
    for (var i = 0; i < 2; i++) {
      promise_list[i].cancel();
    }
  }

  promise = new Promise(function(done, fail) {
    fulfill = function (fulfillmentValue) {
      if (resolved) {return;}
      queue.isFulfilled = true;
      queue.fulfillmentValue = fulfillmentValue;
      resolved = true;
      return done(fulfillmentValue);
    };
    reject = function (rejectedReason) {
      if (resolved) {return;}
      queue.isRejected = true;
      queue.rejectedReason = rejectedReason ;
      resolved = true;
      return fail(rejectedReason);
    };
  }, canceller);

  promise_list.push(resolve());
  promise_list.push(promise_list[0].then(function () {
    promise_list.splice(0, 2);
    if (promise_list.length === 0) {
      fulfill();
    }
  }));

  queue.cancel = function () {
    if (resolved) {return;}
    resolved = true;
    promise.cancel();
    promise.fail(function (rejectedReason) {
      queue.isRejected = true;
      queue.rejectedReason = rejectedReason;
    });
  };
  queue.then = function () {
    return promise.then.apply(promise, arguments);
  };

  queue.push = function(done, fail) {
    var last_promise = promise_list[promise_list.length - 1],
      next_promise;

    if (resolved) {
      throw new ResolvedQueueError();
    }

    next_promise = last_promise.then(done, fail);
    promise_list.push(next_promise);

    // Handle pop
    promise_list.push(next_promise.then(function (fulfillmentValue) {
      promise_list.splice(0, 2);
      if (promise_list.length === 0) {
        fulfill(fulfillmentValue);
      } else {
        return fulfillmentValue;
      }
    }, function (rejectedReason) {
      promise_list.splice(0, 2);
      if (promise_list.length === 0) {
        reject(rejectedReason);
      } else {
        throw rejectedReason;
      }
    }));

    return this;
  };
};

Queue.prototype = Object.create(Promise.prototype);
Queue.prototype.constructor = Queue;

export { Queue, ResolvedQueueError };
